@@ -5,8 +5,8 @@
 # Gibbs Sampler Function
   .gibbs_IFA     <- function(Q, data, iters, N, P, sigma.mu, mu, prop, uni.type,
                              uni.prior, psi.alpha, psi.beta, burnin, thinning, verbose,
-                             sw, epsilon, mu.zero, nu, vrho, adapt, adaptat, b0, b1,
-                             alpha.d1, alpha.d2, beta.d1, beta.d2, scaling, ...) {
+                             sw, epsilon, mu.zero, nu1, nu2, adapt, start.AGS, stop.AGS,
+                             b0, b1, alpha.d1, alpha.d2, beta.d1, beta.d2, scaling, ...) {
 
   # Define & initialise variables
     start.time   <- proc.time()
@@ -14,6 +14,7 @@
     total        <- max(iters)
     if(verbose)     pb     <- utils::txtProgressBar(min=0, max=total, style=3)
     n.store      <- length(iters)
+    AGS.burn     <- total/5L
     Pseq         <- seq_len(P)
     obsnames     <- rownames(data)
     varnames     <- colnames(data)
@@ -39,50 +40,36 @@
     Q.large      <- Q.big  <- FALSE
 
     mu.sigma     <- 1/sigma.mu
-    uni.type     <- switch(uni.type,   unconstrained=,               constrained="constrained", "single")
-    .sim_psi_inv <- switch(uni.type,   constrained=.sim_psi_u1,      single=.sim_psi_c1)
-    .sim_psi_ip  <- switch(uni.prior,  unconstrained=.sim_psi_ipu,   isotropic=.sim_psi_ipc)
-    psi.beta     <- switch(uni.prior,  isotropic=psi.beta[which.max(.ndeci(psi.beta))], psi.beta)
-    uni.shape    <- switch(uni.type,   constrained=N/2 + psi.alpha,  single=(N * P)/2 + psi.alpha)
-    V            <- switch(uni.type,   constrained=P,                single=1)
+    mu.zero      <- as.numeric(mu.zero)
+    uni.type     <- switch(EXPR=uni.type,  unconstrained=,               constrained="constrained", "single")
+    .sim_psi_inv <- switch(EXPR=uni.type,  constrained=.sim_psi_u1,      single=.sim_psi_c1)
+    .sim_psi_ip  <- switch(EXPR=uni.prior, unconstrained=.sim_psi_ipu,   isotropic=.sim_psi_ipc)
+    psi.beta     <- switch(EXPR=uni.prior, isotropic=psi.beta[which.max(.ndeci(psi.beta))], psi.beta)
+    uni.shape    <- switch(EXPR=uni.type,  constrained=N/2 + psi.alpha,  single=(N * P)/2 + psi.alpha)
+    V            <- switch(EXPR=uni.type,  constrained=P,                single=1L)
     eta          <- .sim_eta_p(Q=Q, N=N)
-    phi          <- .sim_phi_p(Q=Q, P=P, nu=nu, rho=vrho)
+    phi          <- .sim_phi_p(Q=Q, P=P, nu1=nu1, nu2=nu2)
     delta        <- c(.sim_delta_p(alpha=alpha.d1, beta=beta.d1), .sim_delta_p(Q=Q, alpha=alpha.d2, beta=beta.d2))
     tau          <- cumprod(delta)
     lmat         <- matrix(vapply(Pseq, function(j) .sim_load_ps(Q=Q, phi=phi[j,], tau=tau), numeric(Q)), nrow=P, byrow=TRUE)
     psi.inv      <- .sim_psi_ip(P=P, psi.alpha=psi.alpha, psi.beta=psi.beta)
-    if(Q   < min(N - 1, Ledermann(P))) {
-      fact       <- try(stats::factanal(data, factors=Q, scores="regression", control=list(nstart=50)), silent=TRUE)
-      if(!inherits(fact, "try-error")) {
-        eta      <- fact$scores
-        lmat     <- unclass(fact$loadings)
-        psi.inv  <- 1/fact$uniquenesses
-      }
-    } else {
-      psi.tmp    <- psi.inv
-      psi.inv[]  <- 1/switch(uni.type, constrained=.col_vars(data, suma=mu), exp(mean(log(.col_vars(data, suma=mu)))))
-      inf.ind    <- is.infinite(psi.inv)
-      psi.inv[inf.ind]     <- psi.tmp[inf.ind]
-    }
+    psi.inv[]    <- 1/switch(EXPR=uni.type, constrained=.col_vars(data, avg=mu), max(.col_vars(data, avg=mu)))
+    max.p        <- (psi.alpha  - 1)/psi.beta
+    inf.ind      <- psi.inv > max(max.p)
+    psi.inv[inf.ind]       <- switch(EXPR=uni.type, constrained=max.p, rep(max.p, P))[inf.ind]
+    rm(max.p, inf.ind)
     sum.data     <- mu * N
-    if(burnin     < 1) {
-      mu.store[,1]         <- mu
-      eta.store[,,1]       <- eta
-      load.store[,,1]      <- lmat
-      psi.store[,1]        <- 1/psi.inv
-      ll.store[1]          <- sum(dmvn(X=data, mu=mu, sigma=tcrossprod(lmat) + diag(1/psi.inv), log=TRUE))
-    }
     init.time    <- proc.time() - start.time
 
   # Iterate
-    for(iter in seq_len(total)[-1]) {
+    for(iter in seq_len(total)) {
       if(verbose && iter    < burnin) utils::setTxtProgressBar(pb, iter)
       storage    <- is.element(iter,  iters)
       Q0         <- Q  > 0
       Q1         <- Q == 1
 
     # Scores & Loadings
-      c.data     <- sweep(data, 2, mu, FUN="-", check.margin=FALSE)
+      c.data     <- sweep(data, 2L, mu, FUN="-", check.margin=FALSE)
       if(Q0) {
         eta      <- .sim_score(N=N, Q=Q, lmat=lmat, psi.inv=psi.inv, c.data=c.data, Q1=Q1)
         lmat     <- matrix(vapply(Pseq, function(j) .sim_load_s(Q=Q, tau=tau, eta=eta, c.data=c.data[,j], Q1=Q1,
@@ -102,21 +89,21 @@
     # Shrinkage
       if(Q0) {
         load.2   <- lmat * lmat
-        phi      <- .sim_phi(Q=Q, P=P, nu=nu, rho=vrho, tau=tau, load.2=load.2)
+        phi      <- .sim_phi(Q=Q, P=P, nu1=nu1, nu2=nu2, tau=tau, load.2=load.2)
 
         sum.term <- colSums2(phi * load.2)
         for(k in seq_len(Q)) {
           delta[k]  <- if(k > 1) .sim_deltak(alpha.d2=alpha.d2, beta.d2=beta.d2, delta.k=delta[k], Q=Q, P=P, k=k,
                        tau.kq=tau[k:Q], sum.term.kq=sum.term[k:Q]) else .sim_delta1(Q=Q, P=P, tau=tau, sum.term=sum.term,
-                       alpha.d1=ifelse(Q1, alpha.d2, alpha.d1), beta.d1=ifelse(Q1, beta.d2, beta.d1), delta.1=delta[1])
+                       alpha.d1=ifelse(Q1, alpha.d2, alpha.d1), beta.d1=ifelse(Q1, beta.d2, beta.d1), delta.1=delta[1L])
           tau       <- cumprod(delta)
         }
       }
 
     # Adaptation
-      if(all(adapt, iter   > adaptat))   {
-        if(stats::runif(1) < ifelse(iter < burnin, 0.5, exp(-b0 - b1  * (iter - adaptat)))) {
-          colvec <- (if(Q0)  colSums(abs(lmat) < epsilon) / P    else stats::runif(1))     >= prop
+      if(adapt   && all(iter >= start.AGS, iter < stop.AGS))      {
+        if(stats::runif(1) < ifelse(iter < AGS.burn, 0.5, exp(-b0 - b1 * (iter - start.AGS)))) {
+          colvec <- (if(Q0)  colSums(abs(lmat)  < epsilon) / P   else stats::runif(1)) >= prop
           numred <- sum(colvec)
           if(numred == 0)  {
             Q    <- Q + 1
@@ -125,14 +112,14 @@
               Q     <- Q.star
             } else {
               eta   <- if(storage) cbind(eta, stats::rnorm(N))   else eta
-              phi   <- cbind(phi,  .rgamma0(n=P, shape=nu, rate=vrho))
+              phi   <- cbind(phi,  .rgamma0(n=P, shape=nu1, rate=nu2))
               delta <- c(delta,    stats::rgamma(n=1, shape=alpha.d2, rate=beta.d2))
               tau   <- cumprod(delta)
               lmat  <- cbind(lmat, stats::rnorm(n=P, mean=0, sd=sqrt(1/(phi[,Q] * tau[Q]))))
             }
           } else if(Q > 0)      {
             nonred  <- colvec  == 0
-            Q       <- max(0, Q - numred)
+            Q       <- max(0L, Q - numred)
             eta     <- if(storage) eta[,nonred, drop=FALSE] else eta
             phi     <- phi[,nonred, drop=FALSE]
             delta   <- delta[nonred]
@@ -142,7 +129,7 @@
         }
       }
 
-      if(Q.big && !Q.large && iter > burnin) {       warning(paste0("Q has exceeded initial number of loadings columns since burnin: consider increasing range.Q from ", Q.star), call.=FALSE)
+      if(Q.big && !Q.large && iter > burnin) {       cat("\n"); warning(paste0("\nQ has exceeded initial number of loadings columns since burnin: consider increasing range.Q from ", Q.star, "\n"), call.=FALSE)
         Q.large  <- TRUE
       }
       if(storage) {
@@ -151,12 +138,12 @@
         psi      <- 1/psi.inv
         post.mu  <- post.mu + mu/n.store
         post.psi <- post.psi + psi/n.store
-        if(sw["mu.sw"])             mu.store[,new.it]              <- mu
-        if(all(sw["s.sw"], Q0))     eta.store[,seq_len(Q),new.it]  <- eta
+        if(sw["mu.sw"])                          mu.store[,new.it] <- mu
+        if(all(sw["s.sw"], Q0))      eta.store[,seq_len(Q),new.it] <- eta
         if(all(sw["l.sw"], Q0))     load.store[,seq_len(Q),new.it] <- lmat
-        if(sw["psi.sw"])            psi.store[,new.it]             <- psi
-                                    Q.store[new.it]                <- as.integer(Q)
-                                    ll.store[new.it]               <- sum(dmvn(X=data, mu=mu, sigma=tcrossprod(lmat) + if(uni) psi else diag(psi), log=TRUE))
+        if(sw["psi.sw"])                        psi.store[,new.it] <- psi
+                                                   Q.store[new.it] <- as.integer(Q)
+                                                  ll.store[new.it] <- sum(dmvn(X=data, mu=mu, sigma=tcrossprod(lmat) + if(uni) psi else diag(psi), log=TRUE))
       }
     }
     if(verbose)     close(pb)
@@ -170,8 +157,8 @@
                          post.mu  = tryCatch(stats::setNames(post.mu,  varnames),            error=function(e) post.mu),
                          post.psi = tryCatch(stats::setNames(post.psi, varnames),            error=function(e) post.psi),
                          ll.store = ll.store,
-                         Q.store  = matrix(Q.store, nrow=1),
+                         Q.store  = matrix(Q.store, nrow=1L),
                          time     = init.time)
     attr(returns, "Q.big") <- Q.large
-    return(returns)
+      return(returns)
   }
