@@ -3,10 +3,10 @@
 ############################################################################
 
 # Gibbs Sampler Function
-  .gibbs_IMIFA       <- function(Q, data, iters, N, P, G, mu.zero, rho, sigma.l, learn.alpha, mu, sw, uni.type,
-                                 uni.prior, sigma.mu, burnin, thinning, a.hyper, psi.alpha, psi.beta, verbose, trunc.G,
-                                 adapt, ind.slice, alpha.d1, discount, alpha.d2, cluster, b0, b1, IM.lab.sw, zeta, tune.zeta,
-                                 nu, vrho, prop, d.hyper, beta.d1, beta.d2, adaptat, epsilon, learn.d, kappa, ...) {
+  .gibbs_IMIFA       <- function(Q, data, iters, N, P, G, mu.zero, rho, sigma.l, learn.alpha, mu, sw, uni.type, uni.prior,
+                                 sigma.mu, burnin, thinning, a.hyper, psi.alpha, psi.beta, verbose, trunc.G, adapt, ind.slice,
+                                 alpha.d1, discount, alpha.d2, cluster, b0, b1, IM.lab.sw, zeta, tune.zeta, rho1, rho2, nu1, nu2,
+                                 cluster.shrink, prop, d.hyper, beta.d1, beta.d2, start.AGS, stop.AGS, epsilon, learn.d, kappa, ...) {
 
   # Define & initialise variables
     start.time       <- proc.time()
@@ -15,6 +15,7 @@
     total            <- max(iters)
     if(verbose)         pb    <- utils::txtProgressBar(min=0, max=total, style=3)
     n.store          <- length(iters)
+    AGS.burn         <- total/5L
     Gs               <- seq_len(G)
     Ts               <- seq_len(trunc.G)
     Ps               <- seq_len(P)
@@ -48,122 +49,88 @@
     err.z            <- z.err <- FALSE
     G.store          <- vector("integer", n.store)
     act.store        <- G.store
+    pi.alpha         <- cluster$pi.alpha
     if(learn.alpha) {
       alpha.store    <- ll.store
-      alpha.shape    <- a.hyper[1]
-      alpha.rate     <- a.hyper[2]
+      alpha.shape    <- a.hyper[1L]
+      alpha.rate     <- a.hyper[2L]
     }
     if(learn.d)     {
       d.store        <- ll.store
-      d.shape1       <- d.hyper[1]
-      d.shape2       <- d.hyper[2]
+      d.shape1       <- d.hyper[1L]
+      d.shape2       <- d.hyper[2L]
       d.rates        <- vector("integer", total)
-      d.unif         <- d.shape1 == 1 & d.shape2 == 1
-    } else d.rates   <- 1
+      d.unif         <- d.shape1 == 1   && d.shape2 == 1
+      .sim_disc_mh   <- if(!learn.alpha && pi.alpha == 0) .sim_d_slab else .sim_d_spike
+    } else d.rates   <- 1L
     MH.step          <- any(discount  > 0, learn.d) && learn.alpha
     if(MH.step)     {
       a.rates        <- vector("integer", total)
-    } else a.rates   <- 1
+    } else a.rates   <- 1L
     if(IM.lab.sw)   {
-      lab.rate       <- matrix(0L, nrow=2, ncol=total)
+      lab.rate       <- matrix(0L, nrow=2L, ncol=total)
     }
-    d.count          <- 0
+    d.count          <- 0L
     avgzeta          <- zeta
     heat             <- tune.zeta$heat
     lambda           <- tune.zeta$lambda
     target           <- tune.zeta$target
-    zeta.tune        <- heat > 0 && tune.zeta$do
+    zeta.tune        <- tune.zeta$do
+    startz           <- tune.zeta$start.zeta
+    stopz            <- tune.zeta$stop.zeta
     mu.sigma         <- 1/sigma.mu
     sig.mu.sqrt      <- sqrt(sigma.mu)
     z                <- cluster$z
-    pi.alpha         <- cluster$pi.alpha
-    one.uni          <- is.element(uni.type, c("constrained", "single"))
-    .sim_psi_inv     <- switch(uni.type,   unconstrained=.sim_psi_uu,   isotropic=.sim_psi_uc,
-                                           constrained=.sim_psi_cu,     single=.sim_psi_cc)
-    .sim_psi_ip      <- switch(uni.prior,  unconstrained=.sim_psi_ipu,  isotropic=.sim_psi_ipc)
-    if(isTRUE(one.uni)) {
-      uni.shape      <- switch(uni.type,   constrained=N/2 + psi.alpha, single=(N * P)/2  + psi.alpha)
-      V              <- switch(uni.type,   constrained=P, single=1)
-    }
-    psi.beta         <- switch(uni.prior,  isotropic=psi.beta[which.max(.ndeci(psi.beta))], psi.beta)
-    pi.prop          <- cluster$pi.prop
     nn               <- tabulate(z, nbins=trunc.G)
     nn0              <- nn > 0
+    nn.ind           <- which(nn > 0)
+    one.uni          <- is.element(uni.type, c("constrained", "single"))
+    .sim_psi_inv     <- switch(EXPR=uni.type,  unconstrained=.sim_psi_uu,   isotropic=.sim_psi_uc,
+                                               constrained=.sim_psi_cu,     single=.sim_psi_cc)
+    .sim_psi_ip      <- switch(EXPR=uni.prior, unconstrained=.sim_psi_ipu,  isotropic=.sim_psi_ipc)
+    if(isTRUE(one.uni)) {
+      uni.shape      <- switch(EXPR=uni.type,  constrained=N/2 + psi.alpha, single=(N * P)/2 + psi.alpha)
+      V              <- switch(EXPR=uni.type,  constrained=P, single=1L)
+    }
+    psi.beta         <- switch(EXPR=uni.prior, isotropic=psi.beta[which.max(.ndeci(psi.beta))], psi.beta)
+    pi.prop          <- cluster$pi.prop
     mu.tmp           <- vapply(seq_len(trunc.G - G), function(g) .sim_mu_p(P=P, sig.mu.sqrt=sig.mu.sqrt, mu.zero=mu.zero), numeric(P))
     mu               <- cbind(mu, if(uni) t(mu.tmp) else mu.tmp)
     eta              <- .sim_eta_p(N=N, Q=Q)
-    phi              <- replicate(trunc.G, .sim_phi_p(Q=Q, P=P, nu=nu, rho=vrho), simplify=FALSE)
-    delta            <- lapply(Ts, function(t) c(.sim_delta_p(alpha=alpha.d1, beta=beta.d1), .sim_delta_p(Q=Q, alpha=alpha.d2, beta=beta.d2)))
+    phi              <- replicate(trunc.G, .sim_phi_p(Q=Q, P=P, nu1=nu1, nu2=nu2), simplify=FALSE)
+    delta            <- replicate(trunc.G, list(c(.sim_delta_p(alpha=alpha.d1, beta=beta.d1), .sim_delta_p(Q=Q, alpha=alpha.d2, beta=beta.d2))))
     tau              <- lapply(delta, cumprod)
-    lmat             <- lapply(Ts, function(t) matrix(vapply(Ps, function(j) .sim_load_ps(Q=Q, phi=phi[[t]][j,], tau=tau[[t]]), numeric(Q)), nrow=P, byrow=TRUE))
-    psi.inv          <- replicate(trunc.G, .sim_psi_ip(P=P, psi.alpha=psi.alpha, psi.beta=psi.beta), simplify="array")
+    if(cluster.shrink)  {
+      sig.store      <- matrix(0L, nrow=trunc.G, ncol=n.store)
+      MGPsig         <- .sim_sigma_p(G=trunc.G, rho1=rho1, rho2=rho2)
+    } else MGPsig    <- rep(1L, trunc.G)
+    lmat             <- lapply(Ts, function(t) matrix(vapply(Ps, function(j) .sim_load_ps(Q=Q, phi=phi[[t]][j,], tau=tau[[t]], sigma=MGPsig[t]), numeric(Q)), nrow=P, byrow=TRUE))
+    if(isTRUE(one.uni)) {
+      psi.inv        <- matrix(.sim_psi_ip(P=P, psi.alpha=psi.alpha, psi.beta=psi.beta), nrow=P, ncol=trunc.G)
+    } else psi.inv   <- replicate(trunc.G, .sim_psi_ip(P=P, psi.alpha=psi.alpha, psi.beta=psi.beta), simplify="array")
     psi.inv          <- if(uni) t(psi.inv) else psi.inv
-    if(Q < min(N - 1, Ledermann(P)))     {
-      for(g in which(nn > P)) {
-        fact         <- try(stats::factanal(data[z == g,, drop=FALSE], factors=Q, scores="regression", control=list(nstart=50)), silent=TRUE)
-        if(!inherits(fact, "try-error")) {
-          eta[z == g,]        <- fact$scores
-          lmat[[g]]           <- unclass(fact$loadings)
-          psi.inv[,g]         <- 1/fact$uniquenesses
-        }
-      }
-    } else {
-      psi.tmp        <- psi.inv
-      if(isTRUE(one.uni)) {
-        psi.inv[,Gs] <- 1/switch(uni.type, constrained=.col_vars(data), exp(mean(log(.col_vars(data)))))
-      } else   {
-        tmp.psi      <- ((nn[nn0] - 1)/(rowsum(data^2,  z) - rowsum(data, z)^2/nn[nn0]))
-        psi.inv[,nn   > 1]    <- tmp.psi[!is.nan(tmp.psi)]
-      }
-      inf.ind        <- is.infinite(psi.inv) | is.nan(psi.inv)
-      psi.inv[inf.ind]        <- psi.tmp[inf.ind]
+    if(isTRUE(one.uni)) {
+      psi.inv[]      <- 1/switch(EXPR=uni.type, constrained=.col_vars(data), max(.col_vars(data)))
+    } else   {
+      tmp.psi        <- (nn[nn0] - 1L)/pmax(rowsum(data^2, z) - rowsum(data, z)^2/nn[nn0], 0L)
+      tmp.psi        <- switch(EXPR=uni.type, unconstrained=t(tmp.psi), matrix(Rfast::rowMaxs(tmp.psi, value=TRUE), nrow=P, ncol=G, byrow=TRUE))
+      psi.inv[,nn     > 1]    <- tmp.psi[!is.nan(tmp.psi)]
+      rm(tmp.psi)
     }
-    psi.inv[psi.inv == 0]     <- colMaxs(psi.inv[,which(psi.inv == 0, arr.ind=TRUE)[,2], drop=FALSE], value=TRUE)
-    index            <- order(pi.prop, decreasing=TRUE)
-    pi.prop          <- pi.prop[index]
-    mu               <- mu[,index, drop=FALSE]
-    phi              <- phi[index]
-    delta            <- delta[index]
-    tau              <- tau[index]
-    lmat             <- lmat[index]
-    psi.inv          <- psi.inv[,index, drop=FALSE]
-    nn               <- nn[index]
-    nn0              <- nn > 0
-    nn.ind           <- which(nn0)
-    G.non            <- length(nn.ind)
-    z                <- factor(z, labels=match(nn.ind, index))
-    z                <- as.integer(levels(z))[z]
+    max.p            <- (psi.alpha  - 1)/psi.beta
+    inf.ind          <- psi.inv  > max(max.p)
+    psi.inv[inf.ind] <- matrix(max.p, nrow=P, ncol=trunc.G)[inf.ind]
+    rm(max.p, inf.ind)
     if(ind.slice) {
-      ksi            <- (1 - rho) * rho^(Ts - 1)
+      ksi            <- (1 - rho)   * rho^(Ts  - 1L)
       log.ksi        <- log(ksi)
-    }
-    slice.logs       <- c(- Inf, 0L)
-    if(burnin         < 1)  {
-      if(sw["mu.sw"])   mu.store[,,1]    <- mu
-      if(sw["s.sw"])    eta.store[,,1]   <- eta
-      if(sw["l.sw"])    load.store[,,,1] <- array(unlist(lmat, use.names=FALSE), dim=c(P, Q, trunc.G))
-      if(sw["psi.sw"])  psi.store[,,1]   <- 1/psi.inv
-      if(sw["pi.sw"])   pi.store[,1]     <- pi.prop
-      z.store[1,]             <- z
-      Q0                      <- Qs > 0
-      sigma                   <- if(uni) lapply(Gs, function(g) as.matrix(1/psi.inv[,g] + if(Q0[g]) tcrossprod(lmat[[g]]) else 0)) else lapply(Gs, function(g) tcrossprod(lmat[[g]]) + diag(1/psi.inv[,g]))
-      log.probs               <- if(uni) vapply(Gs, function(g) stats::dnorm(data, mu[,g], sq_mat(sigma[[g]]), log=TRUE) + log(pi.prop[g]), numeric(N)) else vapply(Gs, function(g, Q=Q0[g]) { sigma <- if(Q) sigma[[g]] else sq_mat(sigma[[g]]); dmvn(data, mu[,g], is.posi_def(sigma, make=TRUE)$X.new, log=TRUE, isChol=!Q) + log(pi.prop[g]) }, numeric(N))
-      ll.store[1]             <- sum(rowLogSumExps(log.probs))
-      Q.store[,1]             <- Qs
-      G.store[1]              <- G.non
-      act.store[1]            <- G
-      if(learn.alpha) {
-        alpha.store[1]        <- pi.alpha
-      }
-      if(learn.d)     {
-        d.store[1]            <- discount
-      }
-    }
+      slinf          <- rep(-Inf, N)
+    } else slinf     <- c(-Inf,  0L)
     init.time        <- proc.time() - start.time
 
   # Iterate
-    for(iter in seq_len(total)[-1]) {
-      if(verbose     && iter   < burnin) utils::setTxtProgressBar(pb, iter)
+    for(iter in seq_len(total))  {
+      if(verbose     && iter     < burnin) utils::setTxtProgressBar(pb, iter)
       storage        <- is.element(iter, iters)
 
     # Mixing Proportions
@@ -171,14 +138,15 @@
       pi.prop        <- .sim_pi_inf(Vs, len=G)
 
     # Re-ordering & Slice Sampler
-      index          <- order(pi.prop[Gs], decreasing=TRUE)
+      index          <- order(pi.prop, decreasing=TRUE)
       prev.prod      <- pi.prop[G]  * (1/Vs[G] - 1)
-      pi.prop[Gs]    <- pi.prop[index]
-      Vs[Gs]         <- Vs[index]
+      pi.prop        <- pi.prop[index]
+      Vs             <- Vs[index]
       mu[,Gs]        <- mu[,index, drop=FALSE]
       phi[Gs]        <- phi[index]
       delta[Gs]      <- delta[index]
       tau[Gs]        <- tau[index]
+      MGPsig[Gs]     <- MGPsig[index]
       lmat[Gs]       <- lmat[index]
       Qs[Gs]         <- Qs[index]
       psi.inv[,Gs]   <- psi.inv[,index, drop=FALSE]
@@ -195,10 +163,10 @@
         G.new        <- sum(min.u   < ksi)
         G.trunc      <- G.new < G.old
         while(G  < G.new && (Vs[G] != 1 || pi.prop[G] != 0)) {
-          newVs      <- .sim_vs_inf(alpha=pi.alpha, discount=discount, len=1, lseq=G + 1)
+          newVs      <- .sim_vs_inf(alpha=pi.alpha, discount=discount, len=1L, lseq=G + 1L)
           Vs         <- c(Vs,      newVs)
           pi.prop    <- c(pi.prop, newVs * prev.prod)
-          G          <- G + 1
+          G          <- G + 1L
           prev.prod  <- pi.prop[G]  * (1/Vs[G] - 1)
         }
         G            <- ifelse(!G.trunc, G, G.new)
@@ -212,21 +180,20 @@
         u.max        <- 1 - min.u
         G.trunc      <- cum.pi > u.max
         while(cum.pi  < u.max && trunc.G > G && (pi.prop[G] != 0 || Vs[G] != 1)) {
-          newVs      <- .sim_vs_inf(alpha=pi.alpha, discount=discount, len=1, lseq=G + 1)
+          newVs      <- .sim_vs_inf(alpha=pi.alpha, discount=discount, len=1L, lseq=G + 1L)
           Vs         <- c(Vs,      newVs)
           newPis     <- newVs  *   prev.prod
           pi.prop    <- c(pi.prop, newPis)
           cum.pi     <- cum.pi +   newPis
           ksi        <- pi.prop
           log.ksi    <- c(log.ksi, log(newPis))
-          G          <- G + 1
+          G          <- G + 1L
           prev.prod  <- pi.prop[G] * (1/Vs[G] - 1)
         }
         G            <- ifelse(G.trunc, which.max(cumsum(pi.prop) > u.max), G)
         Gs           <- seq_len(G)
         if(G.trunc) {
           pi.prop    <- ksi   <-   pi.prop[Gs]
-          log.ksi    <- log.ksi[Gs]
           Vs         <- Vs[Gs]
         }
       }
@@ -235,25 +202,28 @@
       Qgs            <- Qs[Gs]
       Q0             <- Qgs  > 0
       Q1             <- Qgs == 1
-      if(G > 1)     {
+      if(G > 1)       {
         psi          <- 1/psi.inv
-        sigma        <- if(uni) lapply(Gs, function(g) as.matrix(psi[,g] + if(Q0[g]) tcrossprod(lmat[[g]]) else 0)) else lapply(Gs, function(g) tcrossprod(lmat[[g]]) + diag(psi[,g]))
-        if(uni) {
-          log.probs  <- vapply(Gs, function(g) stats::dnorm(data, mu[,g], sq_mat(sigma[[g]]), log=TRUE), numeric(N))
-        } else  {
-          log.probs  <- try(vapply(Gs, function(g, Q=Q0[g]) dmvn(data, mu[,g], if(Q) sigma[[g]] else sq_mat(sigma[[g]]), log=TRUE, isChol=!Q), numeric(N)), silent=TRUE)
-        }
-        if(inherits(log.probs, "try-error")) {
-          log.probs  <- vapply(Gs, function(g, Q=Q0[g]) { sigma <- if(Q) sigma[[g]] else sq_mat(sigma[[g]]); dmvn(data, mu[,g], is.posi_def(sigma, make=TRUE)$X.new, log=TRUE, isChol=!Q) }, numeric(N))
-        }
+        sigma        <- if(uni) lapply(Gs, function(g) as.matrix(psi[,g] + if(Q0[g]) tcrossprod(lmat[[g]]) else 0L)) else lapply(Gs, function(g) tcrossprod(lmat[[g]]) + diag(psi[,g]))
         if(ind.slice) {
-          log.pixi   <- log(pi.prop) - log(ksi[Gs])
-          log.probs  <- vapply(Gs, function(g) slice.logs[1 + (u.slice < ksi[g])] + log.pixi[g], numeric(N)) + log.probs
-        } else   {
-          log.probs  <- vapply(Gs, function(g) slice.logs[1 + (u.slice < ksi[g])], numeric(N)) + log.probs
+          log.pixi   <- log(pi.prop) - log.ksi[Gs]
+          log.probs  <- vapply(Gs, function(g) { slinf[u.slice < ksi[g]] <- log.pixi[g]; slinf }, numeric(N))
+        } else  {
+          log.probs  <- vapply(Gs, function(g) slinf[1L + (u.slice < ksi[g])], numeric(N))
+        }
+        fin.probs    <- is.finite(log.probs)
+        if(uni) {
+          log.probs  <- vapply(Gs, function(g, LP=log.probs[,g], FP=fin.probs[,g]) { LP[FP] <- stats::dnorm(data[FP,], mu[,g], sq_mat(sigma[[g]]), log=TRUE) + LP[FP]; LP }, numeric(N))
+        } else  {
+          probs.log  <- log.probs
+          log.probs  <- try(vapply(Gs, function(g, LP=log.probs[,g], FP=fin.probs[,g]) { LP[FP] <- dmvn(data[FP,], mu[,g], sigma[[g]], log=TRUE,  isChol=!Q) + LP[FP]; LP }, numeric(N)), silent=TRUE)
+          if(z.err   <- inherits(log.probs, "try-error")) {
+           log.probs <- vapply(Gs, function(g, LP=probs.log[,g], FP=fin.probs[,g], Q=Q0[g]) { sigma <- if(Q) is.posi_def(sigma[[g]], make=TRUE)$X.new else sq_mat(sigma[[g]]); LP[FP] <- dmvn(data[FP,], mu[,g], sigma, log=TRUE, isChol=!Q) + LP[FP]; LP }, numeric(N))
+          }
+          rm(probs.log)
         }
         z            <- gumbel_max(probs=log.probs, slice=TRUE)
-      } else   {
+      } else    {
         z            <- rep(1L, N)
       }
       nn             <- tabulate(z, nbins=trunc.G)
@@ -263,22 +233,22 @@
       dat.g          <- lapply(Gs, function(g) data[z == g,, drop=FALSE])
 
     # Scores & Loadings
-      c.data         <- lapply(Gs, function(g) sweep(dat.g[[g]], 2, mu[,g], FUN="-", check.margin=FALSE))
+      c.data         <- lapply(Gs, function(g) sweep(dat.g[[g]], 2L, mu[,g], FUN="-", check.margin=FALSE))
       if(!any(Q0))    {
         eta          <- .empty_mat(nr=N)
-        eta.tmp      <- lapply(Gs, function(g) eta[z == g,, drop=FALSE])
+        eta.tmp      <- lapply(Gs, function(g) eta[z == g,,  drop=FALSE])
         lmat[Gs]     <- replicate(G, .empty_mat(nr=P))
       } else {
         eta.tmp      <- lapply(Gs, function(g) if(all(nn0[g], Q0[g])) .sim_score(N=nn[g], lmat=lmat[[g]], Q=Qs[g], Q1=Q1[g], c.data=c.data[[g]], psi.inv=psi.inv[,g]) else matrix(0, nrow=ifelse(Q0[g], 0, nn[g]), ncol=Qs[g]))
         EtE          <- lapply(Gs, function(g) if(nn0[g]) crossprod(eta.tmp[[g]]))
-        lmat[Gs]     <- lapply(Gs, function(g) matrix(if(all(nn0[g], Q0[g])) vapply(Ps, function(j) .sim_load_s(Q=Qs[g], c.data=c.data[[g]][,j],
-                               Q1=Q1[g], EtE=EtE[[g]], eta=eta.tmp[[g]], psi.inv=psi.inv[,g][j], phi=phi[[g]][j,], tau=tau[[g]]), numeric(Qs[g])) else
-                               vapply(Ps, function(j) .sim_load_ps(Q=Qs[g], phi=phi[[g]][j,], tau=tau[[g]]), numeric(Qs[g])), nrow=P, byrow=TRUE))
+        lmat[Gs]     <- lapply(Gs, function(g) matrix(if(all(nn0[g], Q0[g])) vapply(Ps, function(j) .sim_load_s(Q=Qs[g], c.data=c.data[[g]][,j], Q1=Q1[g],
+                               EtE=EtE[[g]], eta=eta.tmp[[g]], psi.inv=psi.inv[,g][j], phi=phi[[g]][j,], tau=tau[[g]], sigma=MGPsig[g]), numeric(Qs[g])) else
+                               vapply(Ps, function(j) .sim_load_ps(Q=Qs[g], phi=phi[[g]][j,], tau=tau[[g]], sigma=MGPsig[g]), numeric(Qs[g])), nrow=P, byrow=TRUE))
       }
 
     # Uniquenesses
       if(isTRUE(one.uni)) {
-        S.mat        <- lapply(Gs, function(g) { S <- c.data[[g]] - if(Q0[g]) tcrossprod(eta.tmp[[g]], lmat[[g]]) else 0; S * S } )
+        S.mat        <- lapply(Gs, function(g) { S <- c.data[[g]] - if(Q0[g]) tcrossprod(eta.tmp[[g]], lmat[[g]]) else 0L; S * S } )
         psi.inv[,]   <- .sim_psi_inv(uni.shape, psi.beta, S.mat, V)
       } else {
         psi.inv[,Gs] <- vapply(Gs, function(g) if(nn0[g]) .sim_psi_inv(N=nn[g], psi.alpha=psi.alpha, c.data=c.data[[g]], psi.beta=psi.beta, lmat=lmat[[g]],
@@ -295,18 +265,18 @@
     # Shrinkage
       if(all(Q0))     {
         load.2       <- lapply(lmat[Gs], .power2)
-        phi[Gs]      <- lapply(Gs, function(g) if(nn0[g]) .sim_phi(Q=Qs[g], P=P, nu=nu, rho=vrho, tau=tau[[g]],
-                               load.2=load.2[[g]]) else .sim_phi_p(Q=Qs[g], P=P, nu=nu, rho=vrho))
+        phi[Gs]      <- lapply(Gs, function(g) if(nn0[g]) .sim_phi(Q=Qs[g], P=P, nu1=nu1, nu2=nu2, tau=tau[[g]],
+                        load.2=load.2[[g]], sigma=MGPsig[g]) else .sim_phi_p(Q=Qs[g], P=P, nu1=nu1, nu2=nu2))
 
-        sum.terms    <- lapply(Gs, function(g) colSums2(phi[[g]] * load.2[[g]]))
+        sum.terms    <- lapply(Gs, function(g) if(nn0[g]) colSums2(phi[[g]] * load.2[[g]]))
         for(g in Gs)  {
           Qg         <- Qs[g]
           Q1g        <- Q1[g]
           if(nn0[g])  {
             for(k in seq_len(Qg)) {
-              delta[[g]][k]   <- if(k > 1) .sim_deltak(alpha.d2=alpha.d2, beta.d2=beta.d2, delta.k=delta[[g]][k], tau.kq=tau[[g]][k:Qg], P=P,
-                                 Q=Qg, k=k, sum.term.kq=sum.terms[[g]][k:Qg]) else .sim_delta1(Q=Qg, P=P, tau=tau[[g]], sum.term=sum.terms[[g]],
-                                 alpha.d1=ifelse(Q1g, alpha.d2, alpha.d1), beta.d1=ifelse(Q1g, beta.d2, beta.d1), delta.1=delta[[g]][1])
+              delta[[g]][k]   <- if(k > 1) .sim_deltak(alpha.d2=alpha.d2, beta.d2=beta.d2, delta.k=delta[[g]][k], tau.kq=tau[[g]][k:Qg], P=P, Q=Qg,
+                                 k=k, sum.term.kq=sum.terms[[g]][k:Qg], sigma=MGPsig[g]) else .sim_delta1(Q=Qg, P=P, tau=tau[[g]], sum.term=sum.terms[[g]],
+                                 alpha.d1=ifelse(Q1g, alpha.d2, alpha.d1), beta.d1=ifelse(Q1g, beta.d2, beta.d1), delta.1=delta[[g]][1L], sigma=MGPsig[g])
               tau[[g]]        <- cumprod(delta[[g]])
           }
           } else {
@@ -316,24 +286,30 @@
             }
           }
         }
+        if(cluster.shrink)     {
+          nnX                 <- nn0[Gs]
+          nnG0                <- which(nnX)
+          MGPsig[nnG0]        <- .sim_sigma(G=G.non, P=P, Qs=Qs[nnG0], rho1=rho1, rho2=rho2, sum.terms=sum.terms[nnG0], tau=tau[nnG0])
+          MGPsig[which(!nnX)] <- .sim_sigma_p(G=G - G.non, rho1=rho1, rho2=rho2)
+        }
       }
 
     # Adaptation
-      if(all(adapt, iter   > adaptat))   {
-        if(stats::runif(1) < ifelse(iter < burnin, 0.5, exp(-b0 - b1 * (iter - adaptat)))) {
+      if(adapt       && all(iter >= start.AGS, iter < stop.AGS))  {
+        if(stats::runif(1) < ifelse(iter < AGS.burn, 0.5, exp(-b0 - b1 * (iter - start.AGS))))   {
           colvec     <- lapply(nn.ind, function(g) (if(Q0[g]) colSums(abs(lmat[[g]])   < epsilon)/P else stats::runif(1)) >= prop)
           nonred     <- lapply(colvec, .which0)
           numred     <- lengths(colvec)  - lengths(nonred)
           notred     <- numred == 0
           ng.ind     <- seq_along(nn.ind)
           Qs.old     <- Qs[nn0]
-          Qs[nn0]    <- pmax.int(0, vapply(ng.ind, function(h) if(notred[h]) Qs.old[h] + 1 else Qs.old[h] - numred[h], numeric(1L)))
+          Qs[nn0]    <- pmax.int(0L, vapply(ng.ind, function(h) if(notred[h]) Qs.old[h] + 1L else Qs.old[h] - numred[h], numeric(1L)))
           Q.big      <- Qs[nn0] > Q.star
           if((Q.bigs <- any(Q.big))) {
             notred   <- notred & !Q.big
             Qs[nn0][Q.big]    <- Q.star
           }
-          phi[nn0]   <- lapply(nn.ind, function(g, h=which(nn.ind == g)) if(notred[h]) cbind(phi[[g]][,seq_len(Qs.old[h])],  .rgamma0(n=P, shape=nu, rate=vrho)) else phi[[g]][,nonred[[h]], drop=FALSE])
+          phi[nn0]   <- lapply(nn.ind, function(g, h=which(nn.ind == g)) if(notred[h]) cbind(phi[[g]][,seq_len(Qs.old[h])],  .rgamma0(n=P, shape=nu1, rate=nu2)) else phi[[g]][,nonred[[h]], drop=FALSE])
           delta[nn0] <- lapply(nn.ind, function(g, h=which(nn.ind == g)) if(notred[h]) c(delta[[g]][seq_len(Qs.old[h])],     stats::rgamma(n=1, shape=alpha.d2, rate=beta.d2)) else delta[[g]][nonred[[h]]])
           tau[nn0]   <- lapply(delta[nn.ind], cumprod)
           lmat[nn0]  <- lapply(nn.ind, function(g, h=which(nn.ind == g)) if(notred[h]) cbind(lmat[[g]][,seq_len(Qs.old[h])], stats::rnorm(n=P, mean=0, sd=sqrt(1/(phi[[g]][,Qs[g]] * tau[[g]][Qs[g]])))) else lmat[[g]][,nonred[[h]], drop=FALSE])
@@ -353,13 +329,13 @@
                 lmat[[t]]     <- lmat[[t]][,Qmaxseq, drop=FALSE]
               } else  {
                 while(Qt  != Qmax)  {
-                 phi[[t]]     <- cbind(phi[[t]],     .rgamma0(n=P, shape=nu, rate=vrho))
+                 phi[[t]]     <- cbind(phi[[t]],     .rgamma0(n=P, shape=nu1, rate=nu2))
                  delta[[t]]   <- c(delta[[t]],       stats::rgamma(n=1, shape=alpha.d2, rate=beta.d2))
                  tau[[t]]     <- cumprod(delta[[t]])
                  if(store.eta && t %in% Gs)   {
                  eta.tmp[[t]] <- cbind(eta.tmp[[t]], .empty_mat(nc=1))
                  }
-                 Qt  <- Qt + 1
+                 Qt  <- Qt + 1L
                  lmat[[t]]    <- cbind(lmat[[t]],    stats::rnorm(n=P, mean=0, sd=sqrt(1/(phi[[t]][,Qt] * tau[[t]][Qt]))))
                 }
               }
@@ -381,20 +357,21 @@
           a.rate     <- MH.alpha$rate
           if(isTRUE(zeta.tune)) {
             d.count  <- d.count + non.zero.d
-            if(iter   > 100)    {
+            if(iter  >= startz &&
+               iter   < stopz)  {
               zeta   <- .tune_zeta(zeta=zeta, time=d.count, l.rate=MH.alpha$l.prob, heat=heat, target=target, lambda=lambda)
             }
             avgzeta  <- c(avgzeta, zeta)
           }
         } else {
           pi.alpha   <- .sim_alpha_g(alpha=pi.alpha, shape=alpha.shape, rate=alpha.rate, G=G.non, N=N)
-          a.rate     <- 1
+          a.rate     <- 1L
         }
       }
 
     # Discount
       if(learn.d) {
-        MH.disc      <- .sim_disc_mh(discount=discount, alpha=pi.alpha, disc.shape1=d.shape1, disc.shape2=d.shape2, N=N, G=G.non, kappa=kappa, unif=d.unif, nn=nn[nn0])
+        MH.disc      <- .sim_disc_mh(discount=discount, disc.shape1=d.shape1, disc.shape2=d.shape2, G=G.non, unif=d.unif, nn=nn[nn0], alpha=pi.alpha, kappa=kappa)
         discount     <- MH.disc$disc
         d.rate       <- MH.disc$rate
       }
@@ -405,7 +382,7 @@
           move1      <- .lab_move1(nn.ind=nn.ind, pi.prop=pi.prop, nn=nn)
           if((acc1   <- move1$rate1)) {
             sw1      <- move1$sw
-            sw1x     <- c(sw1[2], sw1[1])
+            sw1x     <- c(sw1[2L], sw1[1L])
             nn[sw1]  <- nn[sw1x]
             nn0[sw1] <- nn0[sw1x]
             nn.ind   <- which(nn0)
@@ -414,13 +391,14 @@
             phi[sw1] <- phi[sw1x]
             tau[sw1] <- tau[sw1x]
             Qs[sw1]  <- Qs[sw1x]
+            MGPsig[sw1]       <- MGPsig[sw1x]
             lmat[sw1]         <- lmat[sw1x]
             delta[sw1]        <- delta[sw1x]
             psi.inv[,sw1]     <- psi.inv[,sw1x, drop=FALSE]
             pi.prop[sw1]      <- pi.prop[sw1x]
-            zsw1     <- z == sw1[1]
-            z[z == sw1[2]]    <- sw1[1]
-            z[zsw1]  <- sw1[2]
+            zsw1     <- z == sw1[1L]
+            z[z == sw1[2L]]   <- sw1[1L]
+            z[zsw1]  <- sw1[2L]
             if(all(sw["s.sw"], storage)) {
               eta.tmp[sw1]    <- eta.tmp[sw1x]
               dat.g[sw1]      <- dat.g[sw1x]
@@ -432,7 +410,7 @@
           move2      <- .lab_move2(G=G, Vs=Vs, nn=nn)
           if((acc2   <- move2$rate2)) {
             sw2      <- move2$sw
-            sw2x     <- c(sw2[2], sw2[1])
+            sw2x     <- c(sw2[2L], sw2[1L])
             nn[sw2]  <- nn[sw2x]
             nn0[sw2] <- nn0[sw2x]
             nn.ind   <- which(nn0)
@@ -440,13 +418,14 @@
             phi[sw2] <- phi[sw2x]
             tau[sw2] <- tau[sw2x]
             Qs[sw2]  <- Qs[sw2x]
+            MGPsig[sw2]       <- MGPsig[sw2x]
             lmat[sw2]         <- lmat[sw2x]
             delta[sw2]        <- delta[sw2x]
             psi.inv[,sw2]     <- psi.inv[,sw2x, drop=FALSE]
             pi.prop[sw2]      <- pi.prop[sw2x]
-            zsw2     <- z == sw2[1]
-            z[z == sw2[2]]    <- sw2[1]
-            z[zsw2]  <- sw2[2]
+            zsw2     <- z == sw2[1L]
+            z[z == sw2[2L]]   <- sw2[1L]
+            z[zsw2]  <- sw2[2L]
             if(all(sw["s.sw"], storage)) {
               eta.tmp[sw2]    <- eta.tmp[sw2x]
               dat.g[sw2]      <- dat.g[sw2x]
@@ -456,10 +435,10 @@
         } else  acc2 <- FALSE
       }
 
-      if(Q.bigs && !Q.large   && iter > burnin) {         warning(paste0("Q has exceeded initial number of loadings columns since burnin: consider increasing range.Q from ", Q.star), call.=FALSE)
+      if(Q.bigs && !Q.large   && iter > burnin) {         cat("\n"); warning(paste0("\nQ has exceeded initial number of loadings columns since burnin: consider increasing range.Q from ", Q.star, "\n"), call.=FALSE)
         Q.large      <- TRUE
       }
-      if(z.err  && !err.z) {                              warning("Algorithm may slow due to corrections for Choleski decompositions of non-positive-definite covariance matrices", call.=FALSE)
+      if(z.err  && !err.z) {                              cat("\n"); warning("\nAlgorithm may slow due to corrections for Choleski decompositions of non-positive-definite covariance matrices\n",        call.=FALSE)
         err.z        <- TRUE
       }
       if(MH.step)        a.rates[iter]                 <- a.rate
@@ -468,32 +447,33 @@
       if(storage)    {
         if(verbose)      utils::setTxtProgressBar(pb, iter)
         new.it       <-  which(iters == iter)
-        if(sw["mu.sw"])  mu.store[,,new.it]            <- mu
+        if(sw["mu.sw"])          mu.store[,,new.it]    <-   mu
         if(all(sw["s.sw"],
            any(Q0)))    {
           max.Q      <-  max(Qs)
-          eta.tmp    <-  if(length(unique(Qs)) != 1)      lapply(Gs,       function(g) cbind(eta.tmp[[g]], matrix(0, nrow=nn[g], ncol=max.Q - Qs[g]))) else eta.tmp
+          eta.tmp    <-  if(length(unique(Qs)) != 1)        lapply(Gs,       function(g) cbind(eta.tmp[[g]], matrix(0L, nrow=nn[g], ncol=max.Q - Qs[g]))) else eta.tmp
           q0ng       <-  (!Q0  | Qs[Gs] == 0)   & nn0[Gs]
           if(any(q0ng)) {
-            eta.tmp[q0ng]     <-                          lapply(Gs[q0ng], function(g, x=eta.tmp[[g]]) { row.names(x) <- row.names(dat.g[[g]]); x })
+            eta.tmp[q0ng]     <-                            lapply(Gs[q0ng], function(g, x=eta.tmp[[g]]) { row.names(x) <- row.names(dat.g[[g]]); x })
           }
-                   eta.store[,seq_len(max.Q),new.it]   <- do.call(rbind, eta.tmp)[obsnames,, drop=FALSE]
+                  eta.store[,seq_len(max.Q),new.it]    <-   do.call(rbind, eta.tmp)[obsnames,, drop=FALSE]
         }
         if(sw["l.sw"])  {
           for(g in Gs)  {
             Qseqg    <- seq_len(Qs[g])
-            if(Q0[g])   load.store[,Qseqg,g,new.it]    <- lmat[[g]]
+            if(Q0[g])   load.store[,Qseqg,g,new.it]    <-   lmat[[g]]
           }
         }
-        if(sw["psi.sw"]) psi.store[,,new.it]           <- 1/psi.inv
-        if(sw["pi.sw"])  pi.store[Gs,new.it]           <- pi.prop
-        if(learn.alpha)  alpha.store[new.it]           <- pi.alpha
-        if(learn.d)      d.store[new.it]               <- discount
-                         z.store[new.it,]              <- as.integer(z)
-                         ll.store[new.it]              <- if(G > 1) sum(rowLogSumExps(log.probs)) else sum(dmvn(X=data, mu=mu[,nn.ind], sigma=tcrossprod(lmat[[nn.ind]]) + diag(psi.store[,nn.ind,new.it]), log=TRUE))
-                         Q.store[,new.it]              <- as.integer(Qs)
-                         G.store[new.it]               <- as.integer(G.non)
-                         act.store[new.it]             <- as.integer(G)
+        if(sw["psi.sw"])        psi.store[,,new.it]    <-   1/psi.inv
+        if(sw["pi.sw"])         pi.store[Gs,new.it]    <-   pi.prop
+        if(learn.alpha)         alpha.store[new.it]    <-   pi.alpha
+        if(learn.d)                 d.store[new.it]    <-   discount
+        if(cluster.shrink)       sig.store[,new.it]    <-   MGPsig
+                                   z.store[new.it,]    <-   as.integer(z)
+                                   ll.store[new.it]    <-   if(G > 1) sum(rowLogSumExps(log.probs)) else sum(dmvn(X=data, mu=mu[,nn.ind], sigma=tcrossprod(lmat[[nn.ind]]) + diag(psi.store[,nn.ind,new.it]), log=TRUE))
+                                   Q.store[,new.it]    <-   as.integer(Qs)
+                                    G.store[new.it]    <-   as.integer(G.non)
+                                  act.store[new.it]    <-   as.integer(G)
       }
     }
     if(verbose)         close(pb)
@@ -504,23 +484,24 @@
     load.store       <- if(sw["l.sw"])   tryCatch(load.store[,Qmax,Gmax,, drop=FALSE], error=function(e) load.store)
     psi.store        <- if(sw["psi.sw"]) tryCatch(psi.store[,Gmax,, drop=FALSE],       error=function(e) psi.store)
     pi.store         <- if(sw["pi.sw"])  tryCatch(pi.store[Gmax,, drop=FALSE],         error=function(e) pi.store)
-    returns          <- list(mu        = if(sw["mu.sw"])   tryCatch(provideDimnames(mu.store,  base=list(varnames, "", ""), unique=FALSE), error=function(e) mu.store),
-                             eta       = if(sw["s.sw"])    tryCatch(provideDimnames(tryCatch(as.simple_sparse_array(eta.store),            error=function(e) eta.store),  base=list(obsnames, "", ""),     unique=FALSE), error=function(e) eta.store),
-                             load      = if(sw["l.sw"])    tryCatch(provideDimnames(tryCatch(as.simple_sparse_array(load.store),           error=function(e) load.store), base=list(varnames, "", "", ""), unique=FALSE), error=function(e) load.store),
-                             psi       = if(sw["psi.sw"])  tryCatch(provideDimnames(psi.store, base=list(varnames, "", ""), unique=FALSE), error=function(e) psi.store),
-                             pi.prop   = if(sw["pi.sw"])   pi.store,
-                             alpha     = if(learn.alpha)   alpha.store,
-                             discount  = if(learn.d) {     if(sum(d.store == 0)/n.store > 0.5) as.simple_triplet_matrix(d.store) else d.store },
-                             a.rate    = ifelse(MH.step,   mean(a.rates), a.rates),
-                             d.rate    = ifelse(learn.d,   mean(d.rates), d.rates),
-                             lab.rate  = if(IM.lab.sw)     stats::setNames(rowmeans(lab.rate), c("Move1", "Move2")),
+    returns          <- list(mu        = if(sw["mu.sw"])    tryCatch(provideDimnames(mu.store,  base=list(varnames, "", ""), unique=FALSE), error=function(e) mu.store),
+                             eta       = if(sw["s.sw"])     tryCatch(provideDimnames(tryCatch(as.simple_sparse_array(eta.store),            error=function(e) eta.store),  base=list(obsnames, "", ""),     unique=FALSE), error=function(e) eta.store),
+                             load      = if(sw["l.sw"])     tryCatch(provideDimnames(tryCatch(as.simple_sparse_array(load.store),           error=function(e) load.store), base=list(varnames, "", "", ""), unique=FALSE), error=function(e) load.store),
+                             psi       = if(sw["psi.sw"])   tryCatch(provideDimnames(psi.store, base=list(varnames, "", ""), unique=FALSE), error=function(e) psi.store),
+                             pi.prop   = if(sw["pi.sw"])    pi.store,
+                             sigma     = if(cluster.shrink) sig.store,
+                             alpha     = if(learn.alpha)    alpha.store,
+                             discount  = if(learn.d) {      if(sum(d.store == 0)/n.store > 0.5) as.simple_triplet_matrix(d.store) else d.store },
+                             a.rate    = ifelse(MH.step,    mean(a.rates), a.rates),
+                             d.rate    = ifelse(learn.d,    mean(d.rates), d.rates),
+                             lab.rate  = if(IM.lab.sw)      stats::setNames(rowmeans(lab.rate), c("Move1", "Move2")),
                              z.store   = z.store,
                              ll.store  = ll.store,
                              Q.store   = tryCatch(Q.store[Gmax,, drop=FALSE],          error=function(e) Q.store),
                              G.store   = G.store,
                              act.store = act.store,
-                             avg.zeta  = if(MH.step)       ifelse(zeta.tune, mean(avgzeta), zeta),
+                             avg.zeta  = if(MH.step)        ifelse(zeta.tune, mean(avgzeta), zeta),
                              time      = init.time)
     attr(returns, "Q.big")    <- Q.large
-    return(returns)
+      return(returns)
   }
